# copyright 2003-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.

from os.path import join

from logilab.common.testlib import unittest_main, TestCase

from cubicweb import CW_SOFTWARE_ROOT as BASE
from cubicweb.cwvreg import CWRegistryStore, UnknownProperty
from cubicweb.devtools.apptest_config import ApptestConfiguration
from cubicweb.entity import EntityAdapter


class YesSchema:
    def __contains__(self, something):
        return True


class VRegistryTC(TestCase):
    def setUp(self):
        config = ApptestConfiguration("data", __file__)
        self.vreg = CWRegistryStore(config)
        config.bootstrap_cubes()
        self.vreg.schema = config.load_schema()

    def test_load_interface_based_vobjects(self):
        assert len(self.vreg) == 0

        self.vreg.init_registration([])
        assert "etypes" not in self.vreg

        self.vreg.load_file(
            join(BASE, "entities", "__init__.py"), "cubicweb.entities.__init__"
        )
        assert "etypes" in self.vreg

        self.vreg.initialization_completed()
        assert "etypes" in self.vreg

    def test_load_subinterface_based_appobjects(self):
        assert len(self.vreg) == 0

        # we've to emulate register_objects to add custom MyCard objects
        path = [
            join(BASE, "entities", "__init__.py"),
            join(BASE, "entities", "adapters.py"),
        ]
        filemods = self.vreg.init_registration(path, None)
        assert "etypes" not in self.vreg
        assert "adapters" not in self.vreg

        for filepath, modname in filemods:
            self.vreg.load_file(filepath, modname)

        assert "etypes" in self.vreg
        assert "adapters" in self.vreg

        class CardIDownloadableAdapter(EntityAdapter):
            __regid__ = "IDownloadable"

        self.vreg._loadedmods[__name__] = {}
        self.vreg.register(CardIDownloadableAdapter)
        self.vreg.initialization_completed()

        assert "IDownloadable" in self.vreg["adapters"]
        assert CardIDownloadableAdapter in self.vreg["adapters"]["IDownloadable"]

    def test_properties(self):
        self.vreg.reset()
        self.assertNotIn("system.version.cubicweb", self.vreg["propertydefs"])
        self.assertTrue(self.vreg.property_info("system.version.cubicweb"))
        self.assertRaises(
            UnknownProperty, self.vreg.property_info, "a.non.existent.key"
        )


if __name__ == "__main__":
    unittest_main()



# copyright 2020 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.

from logilab.common.testlib import unittest_main

from cubicweb.devtools.testlib import CubicWebTC
from cubicweb.ext.markdown import markdown_publish


class MarkdownTC(CubicWebTC):
    def context(self, req):
        return req.execute('CWUser X WHERE X login "admin"').get_entity(0, 0)

    def test_basic(self):
        self.assertEqual(markdown_publish(None, "_test_"), "<p><em>test</em></p>")

    def test_urlify(self):
        self.assertEqual(
            markdown_publish(None, "https://www.logilab.fr/"),
            '<p><a href="https://www.logilab.fr/">https://www.logilab.fr/</a></p>',
        )

    def test_table(self):
        self.maxDiff = None
        self.assertEqual(
            markdown_publish(
                None,
                """\
First Header  | Second Header
------------- | -------------
Content Cell  | Content Cell
Content Cell  | Content Cell
""",
            ),
            """\
<table>
<thead>
<tr>
<th>First Header</th>
<th>Second Header</th>
</tr>
</thead>
<tbody>
<tr>
<td>Content Cell</td>
<td>Content Cell</td>
</tr>
<tr>
<td>Content Cell</td>
<td>Content Cell</td>
</tr>
</tbody>
</table>""",
        )


if __name__ == "__main__":
    unittest_main()



# copyright 2015 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""tests for notification hooks"""

from cubicweb.devtools.testlib import CubicWebTC


class NotificationHooksTC(CubicWebTC):
    def test_entity_update(self):
        """Check transaction_data['changes'] filled by "notifentityupdated" hook."""
        with self.admin_access.repo_cnx() as cnx:
            root = cnx.create_entity("Folder", name="a")
            cnx.commit()
            root.cw_set(name="b")
            self.assertIn("changes", cnx.transaction_data)
            self.assertEqual(
                cnx.transaction_data["changes"], {root.eid: {("name", "a", "b")}}
            )


if __name__ == "__main__":
    from logilab.common.testlib import unittest_main

    unittest_main()



# copyright 2003-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""Core hooks: synchronize living session on persistent data changes"""

from cubicweb import UnknownProperty, validation_error, _
from cubicweb.predicates import is_instance
from cubicweb.server import hook


class SyncSessionHook(hook.Hook):
    __abstract__ = True
    category = "syncsession"


class _ChangeSiteWideCWPropertyOp(hook.Operation):
    """Synchronize site wide properties when one has been added/updated"""

    cwprop = None  # make pylint happy

    def postcommit_event(self):
        cwprop = self.cwprop
        if not cwprop.for_user:
            self.cnx.vreg["propertyvalues"][cwprop.pkey] = self.cnx.vreg.typed_value(
                cwprop.pkey, cwprop.value
            )
        # if for_user is set, update is handled by a ChangeUserCWPropertyOp operation


class _DelSiteWideCWPropertyOp(hook.Operation):
    """Synchronize site wide properties when one has been deleted"""

    key = None  # make pylint happy

    def postcommit_event(self):
        self.cnx.vreg["propertyvalues"].pop(self.key, None)


class AddCWPropertyHook(SyncSessionHook):
    __regid__ = "addcwprop"
    __select__ = SyncSessionHook.__select__ & is_instance("CWProperty")
    events = ("after_add_entity",)

    def __call__(self):
        key, value = self.entity.pkey, self.entity.value
        if key.startswith("sources."):
            return
        cnx = self._cw
        try:
            value = cnx.vreg.typed_value(key, value)
        except UnknownProperty:
            msg = _("unknown property key %s")
            raise validation_error(self.entity, {("pkey", "subject"): msg}, (key,))
        except ValueError as ex:
            raise validation_error(self.entity, {("value", "subject"): str(ex)})
        if cnx.user.matching_groups("managers"):
            _ChangeSiteWideCWPropertyOp(cnx, cwprop=self.entity)
        else:
            cnx.add_relation(self.entity.eid, "for_user", cnx.user.eid)


class UpdateCWPropertyHook(AddCWPropertyHook):
    __regid__ = "updatecwprop"
    events = ("after_update_entity",)

    def __call__(self):
        entity = self.entity
        if not ("pkey" in entity.cw_edited or "value" in entity.cw_edited):
            return
        key, value = entity.pkey, entity.value
        if key.startswith("sources."):
            return
        cnx = self._cw
        try:
            value = cnx.vreg.typed_value(key, value)
        except UnknownProperty:
            return
        except ValueError as ex:
            raise validation_error(entity, {("value", "subject"): str(ex)})
        if not entity.for_user:
            _ChangeSiteWideCWPropertyOp(cnx, cwprop=self.entity)


class DeleteCWPropertyHook(AddCWPropertyHook):
    __regid__ = "delcwprop"
    events = ("before_delete_entity",)

    def __call__(self):
        cnx = self._cw
        for eidfrom, rtype, eidto in cnx.transaction_data.get("pendingrelations", ()):
            if rtype == "for_user" and eidfrom == self.entity.eid:
                # not need to sync user specific properties
                break
        else:
            _DelSiteWideCWPropertyOp(cnx, key=self.entity.pkey)


class AddForUserRelationHook(SyncSessionHook):
    __regid__ = "addcwpropforuser"
    __select__ = SyncSessionHook.__select__ & hook.match_rtype("for_user")
    events = ("after_add_relation",)

    def __call__(self):
        cnx = self._cw
        eidfrom = self.eidfrom
        if not cnx.entity_type(eidfrom) == "CWProperty":
            return
        key, value = cnx.execute(
            "Any K,V WHERE P eid %(x)s,P pkey K,P value V", {"x": eidfrom}
        )[0]
        if cnx.vreg.property_info(key)["sitewide"]:
            msg = _("site-wide property can't be set for user")
            raise validation_error(eidfrom, {("for_user", "subject"): msg})


class DelForUserRelationHook(AddForUserRelationHook):
    __regid__ = "delcwpropforuser"
    events = ("after_delete_relation",)

    def __call__(self):
        cnx = self._cw
        cnx.transaction_data.setdefault("pendingrelations", []).append(
            (self.eidfrom, self.rtype, self.eidto)
        )



# copyright 2017 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# copyright 2014-2016 UNLISH S.A.S. (Montpellier, FRANCE), all rights reserved.
#
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.

"""Experimental REST API for CubicWeb using Pyramid."""

import rdflib
from logilab.mtconverter import xml_escape
from pyramid.response import Response
from pyramid.settings import aslist

from cubicweb import rdf
from cubicweb.pyramid.resources import (
    BinaryResource,
    download_context_from_eid,
    download_context_from_identifier,
    rdf_context_from_eid,
    rdf_context_from_identifier,
    RDFResource,
    HTMLResource,
)


def view_entity_as_rdf(context, request):
    graph = rdflib.ConjunctiveGraph()
    rdf.add_entity_to_graph(graph, context.entity)
    rdf_format = rdf.RDF_MIMETYPE_TO_FORMAT[context.mime_type]
    response = Response(graph.serialize(format=rdf_format))
    response.content_type = context.mime_type
    return response


def build_definition_list(e_dict, translator, is_relation):
    parts = []
    for name, value in e_dict.items():
        if not value:
            continue
        if is_relation:
            relation_parts = []
            for entity in value:
                relation_parts.append(
                    f'<li><a href="{entity.absolute_url()}">{entity.dc_title()}</a></li>'
                )
            part = "\n".join(relation_parts)
            parts.append(
                f"""
                <li><b>{translator(name)}</b>:
                    <ul>
                        {part}
                    </ul>
                </li>
                """
            )
        else:
            parts.append(f"<li><b>{translator(name)}</b>: {value}</li>")
    definitions = "\n".join(parts)
    return f"""
    <ul>
    {definitions}
    </ul>
    """


def view_download_entity(context, request):
    _ = request.cw_cnx._
    entity = context.entity
    download_entity = entity.cw_adapt_to("IDownloadable")
    if download_entity is None:
        return Response(status=404)
    response = Response(
        content_type=download_entity.download_content_type(),
        body=download_entity.download_data(),
    )
    response.content_disposition = (
        f'attachment; filename="{download_entity.download_file_name()}"'
    )
    return response


def _render_download_link(eid, entity_download):
    if entity_download is None:
        return None
    return f"<p><a href='/{eid}/download'>download</a></p>"


def view_entity_as_html(context, request):
    _ = request.cw_cnx._
    entity = context.entity
    entity_download = entity.cw_adapt_to("IDownloadable")
    result_dict = {
        "attributes": {},
        "subject_relations": {},
        "object_relations": {},
    }
    for rschema, attribute_schema in entity.e_schema.attribute_definitions():
        # skip binary data
        attribute_name = rschema.type
        if attribute_name == "eid":
            continue
        if attribute_schema.type == "Bytes":
            result_dict["attributes"][
                attribute_name
            ] = "<span style='color: grey'><i>omitted (Binary)</i></span>"
        else:
            result_dict["attributes"][attribute_name] = xml_escape(
                str(getattr(entity, attribute_name))
            )
    for rschema, _d, subject_or_object in entity.e_schema.relation_definitions():
        if rschema.final:
            continue
        relation_name = rschema.type
        if subject_or_object == "subject":
            result_dict["subject_relations"][relation_name] = getattr(
                entity, relation_name
            )
        else:
            result_dict["object_relations"][relation_name] = getattr(
                entity, f"reverse_{relation_name}"
            )
    title = entity.dc_title()
    # we don't use directly translation in f-string to allow to use i18ncubicweb
    attribute_title = _("CWAttribute_plural")
    relation_subject_title = _("Relations (subject)")
    relation_object_title = _("Relations (object)")
    html_content = f"""
    <!DOCTYPE html>
    <html>
    <head>
        <title>{title}</title>
    </head>
    <body>
        <h1>{title}</h1>
        {_render_download_link(entity.eid, entity_download)}
        <h2>{attribute_title}</h2>
        {build_definition_list(result_dict["attributes"], _, is_relation=False)}
        <h2>{relation_subject_title}</h2>
        {build_definition_list(result_dict["subject_relations"], _, is_relation=True)}
        <h2>{relation_object_title}</h2>
        {build_definition_list(result_dict["object_relations"], _, is_relation=True)}
    </body>
    </html>
    """
    response = Response(html_content)
    response.content_type = "text/html"
    return response


def includeme(config):
    cubicweb_includes = aslist(config.registry.settings.get("cubicweb.includes", []))
    mimetypes_to_accept = []
    if "cubicweb.pyramid.rest_api.include_rdf" in cubicweb_includes:
        mimetypes_to_accept = list(rdf.RDF_MIMETYPE_TO_FORMAT.keys())
    if "cubicweb.pyramid.rest_api.include_html" in cubicweb_includes:
        mimetypes_to_accept.append("html/text")

    config.add_route(
        "one_entity",
        "/{etype}/{identifier}",
        factory=rdf_context_from_identifier,
        accept=mimetypes_to_accept,
        request_method="GET",
        match_is_etype_and_identifier=("etype", "identifier"),
    )
    config.add_route(
        "one_entity_eid",
        "/{eid}",
        factory=rdf_context_from_eid,
        accept=mimetypes_to_accept,
        request_method="GET",
        match_is_eid="eid",
    )


def include_rdf(config):
    config.add_view(view_entity_as_rdf, route_name="one_entity", context=RDFResource)
    config.add_view(
        view_entity_as_rdf, route_name="one_entity_eid", context=RDFResource
    )


def include_html(config):
    config.add_view(view_entity_as_html, route_name="one_entity", context=HTMLResource)
    config.add_view(
        view_entity_as_html, route_name="one_entity_eid", context=HTMLResource
    )


def include_download(config):
    config.add_route(
        "one_entity_download",
        "/{etype}/{identifier}/download",
        factory=download_context_from_identifier,
        match_is_etype_and_identifier=("etype", "identifier"),
    )
    config.add_route(
        "one_entity_eid_download",
        "/{eid}/download",
        factory=download_context_from_eid,
        match_is_eid="eid",
    )
    config.add_view(
        view_download_entity, route_name="one_entity_download", context=BinaryResource
    )
    config.add_view(
        view_download_entity,
        route_name="one_entity_eid_download",
        context=BinaryResource,
    )



# copyright 2003-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.

from cubicweb.devtools.testlib import BaseTestCase
from cubicweb.rtags import RelationTags, RelationTagsSet, RelationTagsDict


class RelationTagsTC(BaseTestCase):
    def setUp(self):
        self.rtags = RelationTags(__module__=__name__)
        self.rtags.tag_subject_of(("Societe", "travaille", "*"), "primary")
        self.rtags.tag_subject_of(("*", "evaluee", "*"), "secondary")
        self.rtags.tag_object_of(("*", "tags", "*"), "generated")

    def test_expansion(self):
        self.assertEqual(self.rtags.get("Note", "evaluee", "*", "subject"), "secondary")
        self.assertEqual(
            self.rtags.get("Societe", "travaille", "*", "subject"), "primary"
        )
        self.assertEqual(self.rtags.get("Note", "travaille", "*", "subject"), None)
        self.assertEqual(self.rtags.get("Note", "tags", "*", "subject"), None)
        self.assertEqual(self.rtags.get("*", "tags", "Note", "object"), "generated")
        self.assertEqual(self.rtags.get("Tag", "tags", "*", "object"), "generated")

    def test_expansion_with_parent(self):
        derived_rtags = self.rtags.derive(__name__, None)
        derived_rtags.tag_subject_of(("Societe", "travaille", "*"), "secondary")
        derived_rtags.tag_subject_of(("Note", "evaluee", "*"), "primary")
        self.rtags.tag_object_of(("*", "tags", "*"), "hidden")

        self.assertEqual(
            derived_rtags.get("Note", "evaluee", "*", "subject"), "primary"
        )
        self.assertEqual(
            derived_rtags.get("Societe", "evaluee", "*", "subject"), "secondary"
        )
        self.assertEqual(
            derived_rtags.get("Societe", "travaille", "*", "subject"), "secondary"
        )
        self.assertEqual(derived_rtags.get("Note", "travaille", "*", "subject"), None)
        self.assertEqual(derived_rtags.get("*", "tags", "Note", "object"), "hidden")


class RelationTagsSetTC(BaseTestCase):
    def setUp(self):
        self.rtags = RelationTagsSet(__module__=__name__)
        self.rtags.tag_subject_of(("Societe", "travaille", "*"), "primary")
        self.rtags.tag_subject_of(("*", "travaille", "*"), "secondary")

    def test_expansion(self):
        self.assertEqual(
            self.rtags.get("Societe", "travaille", "*", "subject"),
            {"primary", "secondary"},
        )
        self.assertEqual(
            self.rtags.get("Note", "travaille", "*", "subject"), {"secondary"}
        )
        self.assertEqual(self.rtags.get("Note", "tags", "*", "subject"), set())

    def test_expansion_with_parent(self):
        derived_rtags = self.rtags.derive(__name__, None)
        derived_rtags.tag_subject_of(("Societe", "travaille", "*"), "derived_primary")
        self.assertEqual(
            derived_rtags.get("Societe", "travaille", "*", "subject"),
            {"derived_primary", "secondary"},
        )
        self.assertEqual(
            derived_rtags.get("Note", "travaille", "*", "subject"), {"secondary"}
        )

        derived_rtags.tag_subject_of(("*", "travaille", "*"), "derived_secondary")
        self.assertEqual(
            derived_rtags.get("Societe", "travaille", "*", "subject"),
            {"derived_primary", "derived_secondary"},
        )
        self.assertEqual(
            derived_rtags.get("Note", "travaille", "*", "subject"),
            {"derived_secondary"},
        )

        self.assertEqual(derived_rtags.get("Note", "tags", "*", "subject"), set())


class RelationTagsDictTC(BaseTestCase):
    def setUp(self):
        self.rtags = RelationTagsDict(__module__=__name__)
        self.rtags.tag_subject_of(
            ("Societe", "travaille", "*"), {"key1": "val1", "key2": "val1"}
        )
        self.rtags.tag_subject_of(
            ("*", "travaille", "*"), {"key1": "val0", "key3": "val0"}
        )
        self.rtags.tag_subject_of(("Societe", "travaille", "*"), {"key2": "val2"})

    def test_expansion(self):
        self.assertEqual(
            self.rtags.get("Societe", "travaille", "*", "subject"),
            {"key1": "val1", "key2": "val2", "key3": "val0"},
        )
        self.assertEqual(
            self.rtags.get("Note", "travaille", "*", "subject"),
            {"key1": "val0", "key3": "val0"},
        )
        self.assertEqual(self.rtags.get("Note", "tags", "*", "subject"), {})

        self.rtags.setdefault(("Societe", "travaille", "*", "subject"), "key1", "val4")
        self.rtags.setdefault(("Societe", "travaille", "*", "subject"), "key4", "val4")
        self.assertEqual(
            self.rtags.get("Societe", "travaille", "*", "subject"),
            {"key1": "val1", "key2": "val2", "key3": "val0", "key4": "val4"},
        )

    def test_expansion_with_parent(self):
        derived_rtags = self.rtags.derive(__name__, None)

        derived_rtags.tag_subject_of(("Societe", "travaille", "*"), {"key0": "val0"})
        self.assertEqual(
            derived_rtags.get("Societe", "travaille", "*", "subject"),
            {"key0": "val0", "key1": "val0", "key3": "val0"},
        )
        self.assertEqual(
            derived_rtags.get("Note", "travaille", "*", "subject"),
            {"key1": "val0", "key3": "val0"},
        )
        self.assertEqual(derived_rtags.get("Note", "tags", "*", "subject"), {})

        derived_rtags.tag_subject_of(
            ("*", "travaille", "*"), {"key0": "val00", "key4": "val4"}
        )
        self.assertEqual(
            derived_rtags.get("Societe", "travaille", "*", "subject"),
            {"key0": "val0", "key4": "val4"},
        )
        self.assertEqual(
            derived_rtags.get("Note", "travaille", "*", "subject"),
            {"key0": "val00", "key4": "val4"},
        )


if __name__ == "__main__":
    import unittest

    unittest.main()



# copyright 2003-2010 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""
A RelationTag object is an object which allows to link a configuration
information to a relation definition. For instance, the standard
primary view uses a RelationTag object (uicfg.primaryview_section) to
get the section to display relations.

.. sourcecode:: python

   # display ``entry_of`` relations in the ``relations`` section in the ``BlogEntry`` primary view
   uicfg.primaryview_section.tag_subject_of(('BlogEntry', 'entry_of', '*'),
                                             'relations')

   # hide every relation ``entry_of`` in the ``Blog`` primary view
   uicfg.primaryview_section.tag_object_of(('*', 'entry_of', 'Blog'), 'hidden')

Three primitives are defined:
   * ``tag_subject_of`` tag a relation in the subject's context
   * ``tag_object_of`` tag a relation in the object's context
   * ``tag_attribute`` shortcut for tag_subject_of
"""


import logging

from logilab.common.logging_ext import set_log_methods
from logilab.common.registry import RegistrableInstance, yes


def _ensure_str_key(key):
    return tuple(str(k) for k in key)


def rtags_chain(rtag):
    """Return the rtags chain, starting from the given one, and going back through each parent rtag
    up to the root (i.e. which as no parent).
    """
    while rtag is not None:
        yield rtag
        rtag = rtag._parent


class RegistrableRtags(RegistrableInstance):
    __registry__ = "uicfg"
    __select__ = yes()


class RelationTags(RegistrableRtags):
    """a tag store for full relation definitions :

         (subject type, relation type, object type, tagged)

    allowing to set tags using wildcard (eg '*') as subject type / object type

    This class associates a single tag to each key.
    """

    _allowed_values = None
    # _init expected to be a method (introduced in 3.17), while _initfunc a
    # function given as __init__ argument and kept for bw compat
    _init = _initfunc = None

    def __init__(self, parent=None, __module__=None):
        super().__init__(__module__)
        self._tagdefs = {}
        self._parent = parent
        if parent is not None:
            assert (
                parent.__class__ is self.__class__
            ), f"inconsistent class for parent rtag {parent}"

    def __repr__(self):
        # find a way to have more infos but keep it readable
        # (in error messages in case of an ambiguity for instance)
        return f"<{self.__regid__} {self._short_repr()}>"

    def _short_repr(self):
        # find a way to have more infos but keep it readable
        # (in error messages in case of an ambiguity for instance)
        return "{}@0x{:x}{}".format(
            self.__module__,
            id(self),
            f" derived from {self._parent._short_repr()}" if self._parent else "",
        )

    # dict compat
    def __getitem__(self, key):
        return self.get(*key)

    __contains__ = __getitem__

    def clear(self):
        self._tagdefs.clear()

    def _get_keys(self, stype, rtype, otype, tagged):
        keys = []
        if "*" not in (stype, otype):
            keys.append(("*", rtype, "*", tagged))
        if "*" != stype:
            keys.append(("*", rtype, otype, tagged))
        if "*" != otype:
            keys.append((stype, rtype, "*", tagged))
        keys.append((stype, rtype, otype, tagged))
        return keys

    def init(self, schema, check=True):
        # XXX check existing keys against schema
        if check:
            for (stype, rtype, otype, tagged), value in list(self._tagdefs.items()):
                for ertype in (stype, rtype, otype):
                    if ertype != "*" and ertype not in schema:
                        self.warning(
                            "removing rtag %s: %s, %s undefined in schema",
                            (stype, rtype, otype, tagged),
                            value,
                            ertype,
                        )
                        self.del_rtag(stype, rtype, otype, tagged)
                        break
        if self._parent is None and self._init is not None:
            self.apply(schema, self._init)

    def apply(self, schema, func):
        for eschema in schema.entities():
            if eschema.final:
                continue
            for rschema, tschemas, role in eschema.relation_definitions(True):
                for tschema in tschemas:
                    if role == "subject":
                        sschema, oschema = eschema, tschema
                    else:
                        sschema, oschema = tschema, eschema
                    func(sschema, rschema, oschema, role)

    # rtag declaration api ####################################################

    def derive(self, module, select):
        """Return a derivated of this relation tag, associated to given module and selector.

        This derivated will hold a set of specific rules but delegate to its "parent" relation tags
        for unfound keys.

        >>> class_afs = uicfg.autoform_section.derive(__name__, is_instance('Class'))
        """
        copied = self.__class__(self, __module__=__name__)
        copied.__module__ = module
        copied.__select__ = select
        return copied

    def tag_attribute(self, key, *args, **kwargs):
        key = list(key)
        key.append("*")
        key.append("subject")
        self.tag_relation(key, *args, **kwargs)

    def tag_subject_of(self, key, *args, **kwargs):
        key = list(key)
        key.append("subject")
        self.tag_relation(key, *args, **kwargs)

    def tag_object_of(self, key, *args, **kwargs):
        key = list(key)
        key.append("object")
        self.tag_relation(key, *args, **kwargs)

    def tag_relation(self, key, tag):
        assert len(key) == 4, f"bad key: {list(key)}"
        if self._allowed_values is not None:
            assert (
                tag in self._allowed_values
            ), f"{tag!r} is not an allowed tag (should be in {self._allowed_values})"
        self._tagdefs[_ensure_str_key(key)] = tag
        return tag

    def _tag_etype_attr(self, etype, attr, desttype="*", *args, **kwargs):
        if isinstance(attr, str):
            attr, role = attr, "subject"
        else:
            attr, role = attr
        if role == "subject":
            self.tag_subject_of((etype, attr, desttype), *args, **kwargs)
        else:
            self.tag_object_of((desttype, attr, etype), *args, **kwargs)

    # rtag runtime api ########################################################

    def del_rtag(self, *key):
        del self._tagdefs[key]

    def get(self, *key):
        """Return value for the given key, by looking from the most specific key to the more
        generic (using '*' wildcards). For each key, look into this rtag and its parent rtags.
        """
        for key in reversed(self._get_keys(*key)):
            for rtag in rtags_chain(self):
                try:
                    return rtag._tagdefs[key]
                except KeyError:
                    continue
        return None

    def etype_get(self, etype, rtype, role, ttype="*"):
        if role == "subject":
            return self.get(etype, rtype, ttype, role)
        return self.get(ttype, rtype, etype, role)

    # these are overridden by set_log_methods below
    # only defining here to prevent pylint from complaining
    info = warning = error = critical = exception = debug = lambda msg, *a, **kw: None


class RelationTagsSet(RelationTags):
    """This class associates a set of tags to each key."""

    tag_container_cls = set

    def tag_relation(self, key, tag):
        rtags = self._tagdefs.setdefault(_ensure_str_key(key), self.tag_container_cls())
        rtags.add(tag)
        return rtags

    def get(self, stype, rtype, otype, tagged):
        """Return value for the given key, which is an union of the values found from the most
        specific key to the more generic (using '*' wildcards). For each key, look into this rtag
        and its parent rtags.
        """
        rtags = self.tag_container_cls()
        for key in self._get_keys(stype, rtype, otype, tagged):
            for rtag in rtags_chain(self):
                try:
                    rtags.update(rtag._tagdefs[key])
                    break
                except KeyError:
                    continue
        return rtags


class RelationTagsDict(RelationTagsSet):
    """This class associates a dictionary to each key."""

    tag_container_cls = dict

    def tag_relation(self, key, tag):
        key = _ensure_str_key(key)
        try:
            rtags = self._tagdefs[key]
            rtags.update(tag)
            return rtags
        except KeyError:
            self._tagdefs[key] = tag
            return tag

    def setdefault(self, key, tagkey, tagvalue):
        key = _ensure_str_key(key)
        try:
            rtags = self._tagdefs[key]
            rtags.setdefault(tagkey, tagvalue)
            return rtags
        except KeyError:
            self._tagdefs[key] = {tagkey: tagvalue}
            return self._tagdefs[key]


class RelationTagsBool(RelationTags):
    _allowed_values = frozenset((True, False))


class NoTargetRelationTagsDict(RelationTagsDict):
    @property
    def name(self):
        return self.__class__.name

    # tag_subject_of / tag_object_of issue warning if '*' is not given as target
    # type, while tag_relation handle it silently since it may be used during
    # initialization
    def tag_subject_of(self, key, tag):
        subj, rtype, obj = key
        if obj != "*":
            self.warning(
                "using explict target type in %s.tag_subject_of() "
                'has no effect, use (%s, %s, "*") instead of (%s, %s, %s)',
                self.name,
                subj,
                rtype,
                subj,
                rtype,
                obj,
            )
        super().tag_subject_of((subj, rtype, "*"), tag)

    def tag_object_of(self, key, tag):
        subj, rtype, obj = key
        if subj != "*":
            self.warning(
                "using explict subject type in %s.tag_object_of() "
                'has no effect, use ("*", %s, %s) instead of (%s, %s, %s)',
                self.name,
                rtype,
                obj,
                subj,
                rtype,
                obj,
            )
        super().tag_object_of(("*", rtype, obj), tag)

    def tag_relation(self, key, tag):
        if key[-1] == "subject" and key[-2] != "*":
            if isinstance(key, tuple):
                key = list(key)
            key[-2] = "*"
        elif key[-1] == "object" and key[0] != "*":
            if isinstance(key, tuple):
                key = list(key)
            key[0] = "*"
        super().tag_relation(key, tag)


set_log_methods(RelationTags, logging.getLogger("cubicweb.rtags"))



# copyright 2003-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""CubicWeb is a generic framework to quickly build applications which describes
relations between entitites.
"""


import logging
import os
import pickle
import sys
import warnings
import zlib

from logilab.common.logging_ext import set_log_methods
from yams.constraints import BASE_CONVERTERS, BASE_CHECKERS
from yams.schema import role_name as rname

from cubicweb.__pkginfo__ import version as __version__  # noqa

# make all exceptions accessible from the package
from logilab.common.registry import (  # noqa
    ObjectNotFound,
    NoSelectableObject,
    RegistryNotFound,
)
from yams import ValidationError
from cubicweb._exceptions import *  # noqa

from io import BytesIO

# ignore the pygments UserWarnings
warnings.filterwarnings(
    "ignore",
    category=UserWarning,
    message=".*was already imported",
    module=".*pygments",
)

# pre python 2.7.2 safety
logging.basicConfig()
set_log_methods(sys.modules[__name__], logging.getLogger("cubicweb"))

# this is necessary for i18n devtools test where chdir is done while __path__ is relative, which
# breaks later imports
__path__[0] = os.path.abspath(__path__[0])  # noqa
CW_SOFTWARE_ROOT = __path__[0]  # noqa


# This prefix is often applied to relation types as convention
# indicating a relation is viewed from the object's point of view.
REVERSE_INDICATOR = "reverse_"


# '_' is available to mark internationalized string but should not be used to
# do the actual translation
_ = str


class Binary(BytesIO):
    """class to hold binary data. Use BytesIO to prevent use of unicode data"""

    _allowed_types = (bytes, bytearray, memoryview)

    def __init__(self, buf=b""):
        assert isinstance(
            buf, self._allowed_types
        ), f"Binary objects must use bytes/buffer objects, not {buf.__class__}"
        super().__init__(buf)

    def write(self, data):
        assert isinstance(
            data, self._allowed_types
        ), f"Binary objects must use bytes/buffer objects, not {data.__class__}"
        # don't call super, BytesIO may be an old-style class (on python < 2.7.4)
        BytesIO.write(self, data)

    def to_file(self, fobj):
        """write a binary to disk"""
        pos = self.tell()
        self.seek(0)
        fobj.write(self.read())
        self.seek(pos)

    @staticmethod
    def from_file(filename):
        """read a file and returns its contents in a Binary"""
        binary = Binary()
        with open(filename, "rb") as fobj:
            binary.write(fobj.read())
        binary.seek(0)
        return binary

    def __eq__(self, other):
        if not isinstance(other, Binary):
            return False
        return self.getvalue() == other.getvalue()

    # Binary helpers to store/fetch python objects

    @classmethod
    def zpickle(cls, obj):
        """return a Binary containing a gzipped pickle of obj"""
        retval = cls()
        retval.write(zlib.compress(pickle.dumps(obj, protocol=2)))
        return retval

    def unzpickle(self):
        """decompress and loads the stream before returning it"""
        return pickle.loads(zlib.decompress(self.getvalue()))


def check_password(eschema, value):
    return isinstance(value, (bytes, Binary))


BASE_CHECKERS["Password"] = check_password


def str_or_binary(value):
    if isinstance(value, Binary):
        return value
    return bytes(value)


BASE_CONVERTERS["Password"] = str_or_binary


# use this dictionary to rename entity types while keeping bw compat
ETYPE_NAME_MAP = {}

# XXX cubic web cube migration map. See if it's worth keeping this mecanism
#     to help in cube renaming
CW_MIGRATION_MAP = {}


def neg_role(role):
    if role == "subject":
        return "object"
    return "subject"


def role(obj):
    try:
        return obj.role
    except AttributeError:
        return neg_role(obj.target)


def target(obj):
    try:
        return obj.target
    except AttributeError:
        return neg_role(obj.role)


class CubicWebEventManager:
    """simple event / callback manager.

    Typical usage to register a callback::

      >>> from cubicweb import CW_EVENT_MANAGER
      >>> CW_EVENT_MANAGER.bind('after-registry-reload', mycallback)

    Typical usage to emit an event::

      >>> from cubicweb import CW_EVENT_MANAGER
      >>> CW_EVENT_MANAGER.emit('after-registry-reload')

    emit() accepts an additional context parameter that will be passed
    to the callback if specified (and only in that case)
    """

    def __init__(self):
        self.callbacks = {}

    def bind(self, event, callback, *args, **kwargs):
        self.callbacks.setdefault(event, []).append((callback, args, kwargs))

    def emit(self, event, context=None):
        for callback, args, kwargs in self.callbacks.get(event, ()):
            if context is None:
                callback(*args, **kwargs)
            else:
                callback(context, *args, **kwargs)


CW_EVENT_MANAGER = CubicWebEventManager()


def onevent(event, *args, **kwargs):
    """decorator to ease event / callback binding

    >>> from cubicweb import onevent
    >>> @onevent('before-registry-reload')
    ... def mycallback():
    ...     print 'hello'
    ...
    >>>
    """

    def _decorator(func):
        CW_EVENT_MANAGER.bind(event, func, *args, **kwargs)
        return func

    return _decorator


def validation_error(entity, errors, substitutions=None, i18nvalues=None):
    """easy way to retrieve a :class:`cubicweb.ValidationError` for an entity or eid.

    You may also have 2-tuple as error keys, :func:`yams.role_name` will be
    called automatically for them.

    Messages in errors **should not be translated yet**, though marked for
    internationalization. You may give an additional substition dictionary that
    will be used for interpolation after the translation.
    """
    if substitutions is None:
        # set empty dict else translation won't be done for backward
        # compatibility reason (see ValidationError.translate method)
        substitutions = {}
    for key in list(errors):
        if isinstance(key, tuple):
            errors[rname(*key)] = errors.pop(key)
    return ValidationError(
        getattr(entity, "eid", entity), errors, substitutions, i18nvalues
    )


# exceptions ##################################################################


class ProgrammingError(Exception):
    """Exception raised for errors that are related to the database's operation
    and not necessarily under the control of the programmer, e.g. an unexpected
    disconnect occurs, the data source name is not found, a transaction could
    not be processed, a memory allocation error occurred during processing,
    etc.
    """



# copyright 2017 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# copyright 2014-2016 UNLISH S.A.S. (Montpellier, FRANCE), all rights reserved.
#
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""
Web session when using pyramid
------------------------------

CubicWeb ``CWSession`` entity type so that sessions can be
stored in the database, which allows to run a Cubicweb instance
without having to set up a session storage (like redis or memcache)
solution.

However, for production systems, it is greatly advised to use such a
storage solution for the sessions.

The handling of the sessions is made by pyramid (see the
`pyramid's documentation on sessions`_ for more details).

For example, to set up a redis based session storage, you need the
`pyramid-session-redis`_ package, then you must configure pyramid to
use this backend, by configuring the pyramid configuration file:


.. code-block:: ini

   [main]
   cubicweb.defaults = no # we do not want to load the default cw session handling

   cubicweb.auth.authtkt.session.secret = <secret1>
   cubicweb.auth.authtkt.persistent.secret = <secret2>
   cubicweb.auth.authtkt.session.secure = yes
   cubicweb.auth.authtkt.persistent.secure = yes

   redis.sessions.secret = <secret3>
   redis.sessions.prefix = <my-app>:

   redis.sessions.url = redis://localhost:6379/0

   cubicweb.pyramid.auth = yes

   pyramid.includes =
           pyramid_session_redis


.. Warning:: If you want to be able to log in a CubicWeb application
             served by pyramid on a unsecured stream (typically when
             you start an instance in dev mode using a simple
             ``cubicweb-ctl start -D -linfo myinstance``), you
             **must** set ``cubicweb.auth.authtkt.session.secure`` to
             ``no``.

Secrets
~~~~~~~

There are a number of secrets to configure in ``pyramid.ini``. They
should be different one from each other, as explained in `Pyramid's
documentation`_.

For the record, regarding session handling:

:cubicweb.session.secret: This secret is used to encrypt the session's
   data ID (data themselved are stored in the backend, database or
   redis) when using the integrated (``CWSession`` based) session data
   storage.

:redis.session.secret: This secret is used to encrypt the session's
   data ID (data themselved are stored in the backend, database or
   redis) when using redis as backend.
"""

import logging
import pickle
from contextlib import contextmanager

from pyramid.session import SignedCookieSessionFactory, JSONSerializer, PickleSerializer

from cubicweb import (
    Binary,
    UnknownEid,
)

log = logging.getLogger(__name__)


def logerrors(logger):
    def wrap(fn):
        def newfn(*args, **kw):
            try:
                return fn(*args, **kw)
            except Exception:
                logger.exception(f"Error in {fn.__name__}")

        return newfn

    return wrap


@contextmanager
def unsafe_cnx_context_manager(request):
    """Return a connection for use as a context manager, with security disabled

    If request has an attached connection, its security will be deactived in the context manager's
    scope, else a new internal connection is returned.

    This should be used for read-only queries, not if you intend to commit/rollback some data.
    """
    cnx = request.cw_cnx
    if cnx is None:
        with request.registry["cubicweb.repository"].internal_cnx() as cnx:
            yield cnx
    else:
        with cnx.security_enabled(read=False, write=False):
            yield cnx


class JSONSerializerWithPickleFallback:
    def __init__(self):
        self.json = JSONSerializer()
        self.pickle = PickleSerializer()

    def dumps(self, value):
        # maybe catch serialization errors here and keep using pickle
        # while finding spots in your app that are not storing
        # JSON-serializable objects, falling back to pickle
        return self.json.dumps(value)

    def loads(self, value):
        try:
            return self.json.loads(value)
        except ValueError:
            return self.pickle.loads(value)


def CWSessionFactory(
    secret,
    cookie_name="session",
    max_age=None,
    path="/",
    domain=None,
    secure=False,
    httponly=True,
    set_on_exception=True,
    timeout=1200,
    reissue_time=120,
    hashalg="sha512",
    salt="pyramid.session.",
    serializer=None,
):
    """A pyramid session factory that store session data in the CubicWeb
    database.

    Storage is done with the 'CWSession' entity, which is provided by the
    'pyramid' cube.

    .. warning::

        Although it provides a sane default behavior, this session storage has
        a serious overhead because it uses RQL to access the database.

        Using pure SQL would improve a bit (it is roughly twice faster), but it
        is still pretty slow and thus not an immediate priority.

        It is recommended to use faster session factory
        (pyramid_session_redis_ for example) if you need speed.

    .. _pyramid_session_redis: https://github.com/jvanasco/pyramid_session_redis
    """

    SignedCookieSession = SignedCookieSessionFactory(
        secret,
        cookie_name=cookie_name,
        max_age=max_age,
        path=path,
        domain=domain,
        secure=secure,
        httponly=httponly,
        set_on_exception=set_on_exception,
        timeout=timeout,
        reissue_time=reissue_time,
        hashalg=hashalg,
        salt=salt,
        serializer=serializer if serializer else JSONSerializerWithPickleFallback(),
    )

    class CWSession(SignedCookieSession):
        def __init__(self, request):
            # _set_accessed will be called by the super __init__.
            # Setting _loaded to True inhibates it.
            self._loaded = True

            # the super __init__ will load a single value in the dictionnary,
            # the session id.
            super().__init__(request)

            # Remove the session id from the dict
            self.sessioneid = self.pop("sessioneid", None)
            self.repo = request.registry["cubicweb.repository"]

            # We need to lazy-load only for existing sessions
            self._loaded = self.sessioneid is None

        @logerrors(log)
        def _set_accessed(self, value):
            self._accessed = value

            if self._loaded:
                return

            with unsafe_cnx_context_manager(self.request) as cnx:
                value_rset = cnx.execute(
                    "Any D WHERE X eid %(x)s, X cwsessiondata D", {"x": self.sessioneid}
                )
                value = value_rset[0][0]
                if value:
                    # Use directly dict.update to avoir _set_accessed to be
                    # recursively called
                    dict.update(self, pickle.load(value))

            self._loaded = True

        def _get_accessed(self):
            return self._accessed

        accessed = property(_get_accessed, _set_accessed)

        @logerrors(log)
        def _set_cookie(self, response):
            # Save the value in the database
            data = Binary(pickle.dumps(dict(self)))
            sessioneid = self.sessioneid

            with self.request.registry["cubicweb.repository"].internal_cnx() as cnx:
                if not sessioneid:
                    session = cnx.create_entity("CWSession", cwsessiondata=data)
                    sessioneid = session.eid
                else:
                    try:
                        session = cnx.entity_from_eid(sessioneid)
                    except UnknownEid:
                        # Might occur if CWSession entity got dropped (e.g.
                        # the whole db got recreated) while user's cookie is
                        # still valid. We recreate the CWSession in this case.
                        sessioneid = cnx.create_entity(
                            "CWSession", cwsessiondata=data
                        ).eid
                    else:
                        session.cw_set(cwsessiondata=data)
                cnx.commit()

            # Only if needed actually set the cookie
            if self.new or (
                self._reissue_time is not None
                and self.accessed - self.renewed > self._reissue_time
            ):
                dict.clear(self)
                dict.__setitem__(self, "sessioneid", sessioneid)
                return super()._set_cookie(response)

            return True

    return CWSession


def includeme(config):
    """Activate the CubicWeb session factory.

    It is automatically included by the configuration system, unless the
    following entry is added to the :ref:`pyramid_settings`:

    .. code-block:: ini

        cubicweb.pyramid.session = no

    """
    secret = config.registry.settings["cubicweb.session.secret"]
    session_factory = CWSessionFactory(secret)
    config.set_session_factory(session_factory)



# copyright 2017 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""Configuration for CubicWeb instances on top of a Pyramid application"""

import random
import string

from logilab.common.configuration import merge_options

from cubicweb.cwconfig import CubicWebConfiguration


def get_random_secret_key():
    """Return 50-character secret string"""
    chars = string.ascii_letters + string.digits
    secure_random = random.SystemRandom()

    return "".join([secure_random.choice(chars) for i in range(50)])


class AllInOneConfiguration(CubicWebConfiguration):
    """repository and web instance in the same Pyramid process"""

    name = "all-in-one"
    options = merge_options(
        (
            (
                "profile",
                {
                    "type": "string",
                    "default": None,
                    "help": (
                        "profile code and use the specified file to store stats if this option "
                        "is set"
                    ),
                    "group": "web",
                    "level": 3,
                },
            ),
            (
                "access-control-allow-origin",
                {
                    "type": "csv",
                    "default": (),
                    "help": 'comma-separated list of allowed origin domains or "*" for any domain',
                    "group": "web",
                    "level": 2,
                },
            ),
            (
                "access-control-allow-methods",
                {
                    "type": "csv",
                    "default": (),
                    "help": "comma-separated list of allowed HTTP methods",
                    "group": "web",
                    "level": 2,
                },
            ),
            (
                "access-control-max-age",
                {
                    "type": "int",
                    "default": None,
                    "help": "maximum age of cross-origin resource sharing (in seconds)",
                    "group": "web",
                    "level": 2,
                },
            ),
            (
                "access-control-expose-headers",
                {
                    "type": "csv",
                    "default": (),
                    "help": (
                        "comma-separated list of HTTP headers the application "
                        "declare in response to a preflight request"
                    ),
                    "group": "web",
                    "level": 2,
                },
            ),
            (
                "access-control-allow-headers",
                {
                    "type": "csv",
                    "default": (),
                    "help": (
                        "comma-separated list of HTTP headers the application may set in the "
                        "response"
                    ),
                    "group": "web",
                    "level": 2,
                },
            ),
            (
                "port",
                {
                    "type": "int",
                    "default": None,
                    "help": "http server port number (default to 8080)",
                    "group": "web",
                    "level": 0,
                },
            ),
            (
                "interface",
                {
                    "type": "string",
                    "default": "127.0.0.1",
                    "help": "http server address on which to listen (default to local requests)",
                    "group": "web",
                    "level": 1,
                },
            ),
        )
        + CubicWebConfiguration.options
    )



from unittest import TestCase

from cubicweb import crypto


class CryptoTC(TestCase):
    def test_encrypt_decrypt_roundtrip(self):
        data = {"a": "ah", "b": [1, 2]}
        seed = "s" * 16
        crypted = crypto.encrypt(data, seed)
        decrypted = crypto.decrypt(crypted, seed)
        self.assertEqual(decrypted, data)


if __name__ == "__main__":
    import unittest

    unittest.main()



# copyright 2014 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.

from yams.buildobjs import EntityType, RelationDefinition, ComputedRelation, Int, Float


class Employee(EntityType):
    pass


class employees(RelationDefinition):
    subject = "Company"
    object = "Employee"


class associates(RelationDefinition):
    subject = "Company"
    object = "Employee"


class works_for(ComputedRelation):
    rule = "O employees S, NOT EXISTS (O associates S)"


class Company(EntityType):
    score = Float(formula="Any AVG(NN) WHERE X employees E, N concerns E, N note NN")
    score100 = Float(
        formula="Any AVG(NN) WHERE X employees E, N concerns E, N note100 NN"
    )


class Note(EntityType):
    note = Int()
    note100 = Int(formula="Any N*100 WHERE X note N")


class concerns(RelationDefinition):
    subject = "Note"
    object = "Employee"


class whatever(ComputedRelation):
    rule = "S employees E, O associates E"


class renamed(ComputedRelation):
    rule = "S employees E, O concerns E"


class perm_changes(ComputedRelation):
    __permissions__ = {"read": ("managers",)}
    rule = "S employees E, O concerns E"



# copyright 2003 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""unit tests for modules cubicweb.server.rqlannotation"""
from cubicweb.devtools import get_test_db_handler
from cubicweb.devtools.apptest_config import ApptestConfiguration
from cubicweb.devtools.repotest import BaseQuerierTC


class RQLAnnotatorTC(BaseQuerierTC):
    def setUp(self):
        handler = get_test_db_handler(ApptestConfiguration("data", __file__))
        handler.build_db_cache()
        repo, _cnx = handler.get_repo_and_cnx()
        self.__class__.repo = repo
        super().setUp()

    def test_0_1(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(
                cnx,
                "Any SEN,RN,OEN WHERE X from_entity SE, "
                "SE eid 44, X relation_type R, R eid 139, "
                "X to_entity OE, OE eid 42, R name RN, SE name SEN, "
                "OE name OEN",
            )
            self.assertEqual(rqlst.defined_vars["SE"]._q_invariant, False)
            self.assertEqual(rqlst.defined_vars["OE"]._q_invariant, False)
            self.assertEqual(rqlst.defined_vars["R"]._q_invariant, False)
            self.assertEqual(rqlst.defined_vars["SE"].stinfo["attrvar"], None)
            self.assertEqual(rqlst.defined_vars["OE"].stinfo["attrvar"], None)
            self.assertEqual(rqlst.defined_vars["R"].stinfo["attrvar"], None)

    def test_0_2(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(
                cnx,
                "Any O WHERE NOT S ecrit_par O, S eid 1, " "S inline1 P, O inline2 P",
            )
            self.assertEqual(rqlst.defined_vars["P"]._q_invariant, True)
            self.assertEqual(rqlst.defined_vars["O"].stinfo["attrvar"], None)

    def test_0_4(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(
                cnx, "Any A,B,C WHERE A eid 12,A comment B, " "A ?wf_info_for C"
            )
            self.assertEqual(rqlst.defined_vars["A"]._q_invariant, False)
            self.assertTrue(rqlst.defined_vars["B"].stinfo["attrvar"])
            self.assertEqual(rqlst.defined_vars["C"]._q_invariant, False)
            self.assertEqual(
                rqlst.solutions,
                [
                    {"A": "TrInfo", "B": "String", "C": "Affaire"},
                    {"A": "TrInfo", "B": "String", "C": "CWUser"},
                    {"A": "TrInfo", "B": "String", "C": "Note"},
                ],
            )

    def test_0_5(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(cnx, "Any P WHERE N ecrit_par P, N eid 0")
            self.assertEqual(rqlst.defined_vars["N"]._q_invariant, False)
            self.assertEqual(rqlst.defined_vars["P"]._q_invariant, True)

    def test_0_6(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(cnx, "Any P WHERE NOT N ecrit_par P, N eid 512")
            self.assertEqual(rqlst.defined_vars["P"]._q_invariant, False)

    def test_0_7(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(
                cnx, "Personne X,Y where X nom NX, " "Y nom NX, X eid XE, not Y eid XE"
            )
            self.assertEqual(rqlst.defined_vars["X"]._q_invariant, False)
            self.assertEqual(rqlst.defined_vars["Y"]._q_invariant, False)
            self.assertTrue(rqlst.defined_vars["XE"].stinfo["attrvar"])

    def test_0_8(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(cnx, "Any P WHERE X eid 0, NOT X connait P")
            self.assertEqual(rqlst.defined_vars["P"]._q_invariant, False)
            self.assertEqual(len(rqlst.solutions), 1, rqlst.solutions)

    def test_0_10(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(cnx, "Any X WHERE X concerne Y, Y is Note")
            self.assertEqual(rqlst.defined_vars["X"]._q_invariant, True)
            self.assertEqual(rqlst.defined_vars["Y"]._q_invariant, False)

    def test_0_11(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(cnx, "Any X WHERE X todo_by Y, X is Affaire")
            self.assertEqual(rqlst.defined_vars["X"]._q_invariant, False)
            self.assertEqual(rqlst.defined_vars["Y"]._q_invariant, True)

    def test_0_12(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(
                cnx, "Personne P WHERE P concerne A, " 'A concerne S, S nom "Logilab"'
            )
            self.assertEqual(rqlst.defined_vars["P"]._q_invariant, True)
            self.assertEqual(rqlst.defined_vars["A"]._q_invariant, True)
            self.assertEqual(rqlst.defined_vars["S"]._q_invariant, False)

    def test_1_0(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(
                cnx, "Any X,Y WHERE X created_by Y, " "X eid 5, NOT Y eid 6"
            )
            self.assertEqual(rqlst.defined_vars["Y"]._q_invariant, True)

    def test_1_1(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(
                cnx, "Any X,Y WHERE X created_by Y, X eid 5, " "NOT Y eid IN (6,7)"
            )
            self.assertEqual(rqlst.defined_vars["Y"]._q_invariant, True)

    def test_2(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(cnx, "Any X WHERE X identity Y, Y eid 1")
            self.assertEqual(rqlst.defined_vars["X"]._q_invariant, False)

    def test_7(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(
                cnx, "Personne X,Y where X nom NX, Y nom NX, " "X eid XE, not Y eid XE"
            )
            self.assertEqual(rqlst.defined_vars["X"]._q_invariant, False)
            self.assertEqual(rqlst.defined_vars["Y"]._q_invariant, False)

    def test_8(self):
        with self.admin_access.cnx() as cnx:
            # DISTINCT Any P WHERE P require_group %(g)s,
            # NOT %(u)s has_group_permission P, P is CWPermission
            rqlst = self._prepare(
                cnx,
                "DISTINCT Any X WHERE A concerne X, "
                "NOT N migrated_from X, "
                "X is Note, N eid 1",
            )
            self.assertEqual(rqlst.defined_vars["X"]._q_invariant, False)

    def test_diff_scope_identity_deamb(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(
                cnx,
                "Any X WHERE X concerne Y, Y is Note, "
                "EXISTS(Y identity Z, Z migrated_from N)",
            )
            self.assertEqual(rqlst.defined_vars["Z"]._q_invariant, True)
            self.assertEqual(rqlst.defined_vars["Y"]._q_invariant, True)

    def test_optional_inlined(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(cnx, "Any X,S where X from_state S?")
            self.assertEqual(rqlst.defined_vars["X"]._q_invariant, False)
            self.assertEqual(rqlst.defined_vars["S"]._q_invariant, True)

    def test_optional_inlined_2(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(cnx, "Any N,A WHERE N? inline1 A")
            self.assertEqual(rqlst.defined_vars["N"]._q_invariant, False)
            self.assertEqual(rqlst.defined_vars["A"]._q_invariant, False)

    def test_optional_1(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(cnx, "Any X,S WHERE X travaille S?")
            self.assertEqual(rqlst.defined_vars["X"]._q_invariant, False)
            self.assertEqual(rqlst.defined_vars["S"]._q_invariant, True)

    def test_greater_eid(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(cnx, "Any X WHERE X eid > 5")
            self.assertEqual(rqlst.defined_vars["X"]._q_invariant, True)

    def test_greater_eid_typed(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(cnx, "Any X WHERE X eid > 5, X is Note")
            self.assertEqual(rqlst.defined_vars["X"]._q_invariant, False)

    def test_max_eid(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(cnx, "Any MAX(X)")
            self.assertEqual(rqlst.defined_vars["X"]._q_invariant, True)

    def test_max_eid_typed(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(cnx, "Any MAX(X) WHERE X is Note")
            self.assertEqual(rqlst.defined_vars["X"]._q_invariant, False)

    def test_all_entities(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(cnx, "Any X")
            self.assertEqual(rqlst.defined_vars["X"]._q_invariant, True)

    def test_all_typed_entity(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(cnx, "Any X WHERE X is Note")
            self.assertEqual(rqlst.defined_vars["X"]._q_invariant, False)

    def test_has_text_1(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(cnx, 'Any X WHERE X has_text "toto tata"')
            self.assertEqual(rqlst.defined_vars["X"]._q_invariant, True)
            self.assertEqual(
                rqlst.defined_vars["X"].stinfo["principal"].r_type, "has_text"
            )

    def test_has_text_2(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(
                cnx, "Any X WHERE X is Personne, " 'X has_text "coucou"'
            )
            self.assertEqual(rqlst.defined_vars["X"]._q_invariant, True)
            self.assertEqual(
                rqlst.defined_vars["X"].stinfo["principal"].r_type, "has_text"
            )

    def test_not_relation_1(self):
        with self.admin_access.cnx() as cnx:
            # P can't be invariant since deambiguification caused by "NOT X require_permission P"
            # is not considered by generated sql (NOT EXISTS(...))
            rqlst = self._prepare(
                cnx, "Any P,G WHERE P require_group G, " "NOT X require_permission P"
            )
            self.assertEqual(rqlst.defined_vars["P"]._q_invariant, False)
            self.assertEqual(rqlst.defined_vars["G"]._q_invariant, True)
            self.assertEqual(rqlst.defined_vars["X"]._q_invariant, True)

    def test_not_relation_2(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(
                cnx, "TrInfo X WHERE X eid 2, " "NOT X from_state Y, Y is State"
            )
            self.assertEqual(rqlst.defined_vars["Y"]._q_invariant, True)
            self.assertEqual(rqlst.defined_vars["X"]._q_invariant, False)

    def test_not_relation_3(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(cnx, "Any X, Y WHERE X eid 1, Y eid in (2, 3)")
            self.assertEqual(rqlst.defined_vars["Y"]._q_invariant, False)

    def test_not_relation_4_1(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(cnx, "Note X WHERE NOT Y evaluee X")
            self.assertEqual(rqlst.defined_vars["X"]._q_invariant, False)
            self.assertEqual(rqlst.defined_vars["Y"]._q_invariant, True)

    def test_not_relation_4_2(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(cnx, "Any X WHERE NOT Y evaluee X")
            self.assertEqual(rqlst.defined_vars["X"]._q_invariant, False)
            self.assertEqual(rqlst.defined_vars["Y"]._q_invariant, True)

    def test_not_relation_4_3(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(cnx, "Any Y WHERE NOT Y evaluee X")
            self.assertEqual(rqlst.defined_vars["X"]._q_invariant, True)
            self.assertEqual(rqlst.defined_vars["Y"]._q_invariant, False)

    def test_not_relation_4_4(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(cnx, "Any X WHERE NOT Y evaluee X, Y is CWUser")
            self.assertEqual(rqlst.defined_vars["X"]._q_invariant, False)
            self.assertEqual(rqlst.defined_vars["Y"]._q_invariant, False)

    def test_not_relation_4_5(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(
                cnx, f"Any X WHERE NOT Y evaluee X, Y eid {self.ueid}, X is Note"
            )
            self.assertEqual(rqlst.defined_vars["X"]._q_invariant, False)
            self.assertEqual(rqlst.solutions, [{"X": "Note"}])

    def test_not_relation_5_1(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(
                cnx,
                'Any X,Y WHERE X name "CWGroup", '
                "Y eid IN(1, 2, 3), NOT X read_permission Y",
            )
            self.assertEqual(rqlst.defined_vars["X"]._q_invariant, False)
            self.assertEqual(rqlst.defined_vars["Y"]._q_invariant, False)

    def test_not_relation_5_2(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(
                cnx,
                'DISTINCT Any X,Y WHERE X name "CWGroup", '
                "Y eid IN(1, 2, 3), NOT X read_permission Y",
            )
            self.assertEqual(rqlst.defined_vars["X"]._q_invariant, False)
            self.assertEqual(rqlst.defined_vars["Y"]._q_invariant, False)

    def test_not_relation_6(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(cnx, "Personne P where NOT P concerne A")
            self.assertEqual(rqlst.defined_vars["P"]._q_invariant, False)
            self.assertEqual(rqlst.defined_vars["A"]._q_invariant, True)

    def test_not_relation_7(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(
                cnx,
                "Any K,V WHERE P is CWProperty, "
                "P pkey K, P value V, NOT P for_user U",
            )
            self.assertEqual(rqlst.defined_vars["P"]._q_invariant, False)
            self.assertEqual(rqlst.defined_vars["U"]._q_invariant, True)

    def test_exists_1(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(
                cnx, "Any U WHERE U eid IN (1,2), EXISTS(X owned_by U)"
            )
            self.assertEqual(rqlst.defined_vars["U"]._q_invariant, False)
            self.assertEqual(rqlst.defined_vars["X"]._q_invariant, True)

    def test_exists_2(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(
                cnx, "Any U WHERE EXISTS(U eid IN (1,2), X owned_by U)"
            )
            self.assertEqual(rqlst.defined_vars["U"]._q_invariant, False)
            self.assertEqual(rqlst.defined_vars["X"]._q_invariant, True)

    def test_exists_3(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(
                cnx, "Any U WHERE EXISTS(X owned_by U, X bookmarked_by U)"
            )
            self.assertEqual(rqlst.defined_vars["U"]._q_invariant, False)
            self.assertEqual(rqlst.defined_vars["X"]._q_invariant, True)

    def test_exists_4(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(
                cnx,
                'Any X,Y WHERE X name "CWGroup", '
                "Y eid IN(1, 2, 3), EXISTS(X read_permission Y)",
            )
            self.assertEqual(rqlst.defined_vars["X"]._q_invariant, False)
            self.assertEqual(rqlst.defined_vars["Y"]._q_invariant, False)

    def test_exists_5(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(
                cnx,
                'DISTINCT Any X,Y WHERE X name "CWGroup", '
                "Y eid IN(1, 2, 3), EXISTS(X read_permission Y)",
            )
            self.assertEqual(rqlst.defined_vars["X"]._q_invariant, False)
            self.assertEqual(rqlst.defined_vars["Y"]._q_invariant, True)

    def test_not_exists_1(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(
                cnx, "Any U WHERE NOT EXISTS(X owned_by U, " "X bookmarked_by U)"
            )
            self.assertEqual(rqlst.defined_vars["U"]._q_invariant, False)
            self.assertEqual(rqlst.defined_vars["X"]._q_invariant, True)

    def test_not_exists_2(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(
                cnx,
                'Any X,Y WHERE X name "CWGroup", '
                "Y eid IN(1, 2, 3), NOT EXISTS(X read_permission Y)",
            )
            self.assertEqual(rqlst.defined_vars["Y"]._q_invariant, False)

    def test_not_exists_distinct_1(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(
                cnx,
                'DISTINCT Any X,Y WHERE X name "CWGroup", '
                "Y eid IN(1, 2, 3), NOT EXISTS(X read_permission Y)",
            )
            self.assertEqual(rqlst.defined_vars["Y"]._q_invariant, False)

    def test_or_1(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(
                cnx, "Any X WHERE X concerne B OR " "C concerne X, B eid 12, C eid 13"
            )
            self.assertEqual(rqlst.defined_vars["X"]._q_invariant, False)

    def test_or_2(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(
                cnx,
                "Any X WHERE X created_by U, X concerne B OR "
                "C concerne X, B eid 12, C eid 13",
            )
            self.assertEqual(rqlst.defined_vars["X"]._q_invariant, True)
            self.assertEqual(rqlst.defined_vars["U"]._q_invariant, True)
            self.assertEqual(
                rqlst.defined_vars["X"].stinfo["principal"].r_type, "created_by"
            )

    def test_or_3(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(cnx, "Any N WHERE A evaluee N or EXISTS(N todo_by U)")
            self.assertEqual(rqlst.defined_vars["N"]._q_invariant, False)
            self.assertEqual(rqlst.defined_vars["A"]._q_invariant, True)
            self.assertEqual(rqlst.defined_vars["U"]._q_invariant, True)

    def test_or_exists_1(self):
        with self.admin_access.cnx() as cnx:
            # query generated by security rewriting
            rqlst = self._prepare(
                cnx,
                'DISTINCT Any A,S WHERE A is Affaire, S nom "chouette", '
                "S is IN(Division, Societe, SubDivision),"
                "(EXISTS(A owned_by D)) "
                "OR ((((EXISTS(E concerne C?, C owned_by D, A identity E, "
                "              C is Note, E is Affaire)) "
                "OR (EXISTS(I concerne H?, H owned_by D, H is Societe, "
                "           A identity I, I is Affaire))) "
                "OR (EXISTS(J concerne G?, G owned_by D, G is SubDivision, "
                "           A identity J, J is Affaire))) "
                "OR (EXISTS(K concerne F?, F owned_by D, F is Division, "
                "           A identity K, K is Affaire)))",
            )
            self.assertEqual(rqlst.defined_vars["A"]._q_invariant, False)
            self.assertEqual(rqlst.defined_vars["S"]._q_invariant, False)

    def test_or_exists_2(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(
                cnx,
                'Any U WHERE EXISTS(U in_group G, G name "managers") OR '
                "EXISTS(X owned_by U, X bookmarked_by U)",
            )
            self.assertEqual(rqlst.defined_vars["U"]._q_invariant, False)
            self.assertEqual(rqlst.defined_vars["G"]._q_invariant, False)
            self.assertEqual(rqlst.defined_vars["X"]._q_invariant, True)

    def test_or_exists_3(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(
                cnx,
                "Any COUNT(S),CS GROUPBY CS ORDERBY 1 DESC LIMIT 10 "
                "WHERE C is Societe, S concerne C, C nom CS, "
                "(EXISTS(S owned_by D)) "
                'OR (EXISTS(S documented_by N, N title "published"))',
            )
            self.assertEqual(rqlst.defined_vars["S"]._q_invariant, True)
            rqlst = self._prepare(
                cnx,
                "Any COUNT(S),CS GROUPBY CS ORDERBY 1 DESC LIMIT 10 "
                "WHERE S is Affaire, C is Societe, S concerne C, C nom CS, "
                "(EXISTS(S owned_by D)) "
                'OR (EXISTS(S documented_by N, N title "published"))',
            )
            self.assertEqual(rqlst.defined_vars["S"]._q_invariant, True)

    def test_nonregr_ambiguity(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(cnx, "Note N WHERE N attachment F")
            # N may be an image as well, not invariant
            self.assertEqual(rqlst.defined_vars["N"]._q_invariant, False)
            self.assertEqual(rqlst.defined_vars["F"]._q_invariant, True)

    def test_nonregr_ambiguity_2(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(
                cnx,
                'Any S,SN WHERE X has_text "tot", '
                "X in_state S, S name SN, X is CWUser",
            )
            # X use has_text but should not be invariant as ambiguous, and has_text
            # may not be its principal
            self.assertEqual(rqlst.defined_vars["X"]._q_invariant, False)
            self.assertEqual(rqlst.defined_vars["S"]._q_invariant, False)

    def test_remove_from_deleted_source_1(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(cnx, "Note X WHERE X eid 999998, NOT X cw_source Y")
            self.assertNotIn("X", rqlst.defined_vars)  # simplified
            self.assertEqual(rqlst.defined_vars["Y"]._q_invariant, True)

    def test_remove_from_deleted_source_2(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(
                cnx, "Note X WHERE X eid IN (999998, 999999), NOT X cw_source Y"
            )
            self.assertEqual(rqlst.defined_vars["X"]._q_invariant, False)
            self.assertEqual(rqlst.defined_vars["Y"]._q_invariant, True)

    def test_has_text_security_cache_bug(self):
        with self.admin_access.cnx() as cnx:
            rqlst = self._prepare(
                cnx,
                'Any X WHERE X has_text "toto" WITH X BEING '
                "(Any C WHERE C is Societe, C nom CS)",
            )
            self.assertTrue(rqlst.parent.has_text_query)


if __name__ == "__main__":
    import unittest

    unittest.main()



# copyright 2017 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# copyright 2014-2016 UNLISH S.A.S. (Montpellier, FRANCE), all rights reserved.
#
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.

"""Contains predicates used in Pyramid views.
"""

from cubicweb._exceptions import UnknownEid
from cubicweb.pyramid.resources import ETypeResource, negociate_mime_type


class MatchIsETypePredicate:
    """A predicate that match if a given etype exist in schema."""

    def __init__(self, matchname, config):
        self.matchname = matchname

    def text(self):
        return f"match_is_etype = {self.matchname}"

    phash = text

    def __call__(self, info, request):
        return (
            info["match"][self.matchname].lower()
            in request.registry["cubicweb.registry"].case_insensitive_etypes
        )


class MatchIsEIDPredicate:
    """A predicate that match if a given eid exist in the database."""

    def __init__(self, matchname, config):
        self.matchname = matchname

    def text(self):
        return f"match_is_eid = {self.matchname}"

    phash = text

    def __call__(self, info, request):
        try:
            eid = int(info["match"][self.matchname])
        except ValueError:
            return False

        if request.cw_cnx is None:
            return False

        try:
            request.cw_cnx.entity_from_eid(eid)
        except UnknownEid:
            return False
        return True


class MatchIsETypeAndIdentifierPredicate:
    """A predicate which verifiy if the given identifier match one of the rest_attr
    for the given entity type"""

    def __init__(self, matchnames, config):
        self.match_etype, self.match_identifier = matchnames

    def text(self):
        return (
            "match_is_etype_and_identifier ="
            f" {self.match_etype}/{self.match_identifier}"
        )

    phash = text

    def __call__(self, info, request):
        etype = info["match"]["etype"]
        identifier = info["match"]["identifier"]
        try:
            etype_resource = ETypeResource(request, etype)
            resource = etype_resource[identifier]
            entity = resource.rset.one()
            if entity:
                return True
        except Exception:
            pass
        return False


class IsRequestMimeTypeRDFFormat:
    def __init__(self, val, info):
        self.available_rdf_mimetype = val

    def text(self):
        return f"is_request_mimetype_rdf_format = {self.available_rdf_mimetype}"

    phash = text

    def __call__(self, info, request):
        mime_type = negociate_mime_type(request, self.available_rdf_mimetype)
        return mime_type is not None


def includeme(config):
    config.add_route_predicate("match_is_etype", MatchIsETypePredicate)
    config.add_route_predicate("match_is_eid", MatchIsEIDPredicate)
    config.add_route_predicate(
        "match_is_etype_and_identifier", MatchIsETypeAndIdentifierPredicate
    )
    config.add_route_predicate(
        "is_request_mimetype_rdf_format", IsRequestMimeTypeRDFFormat
    )



# copyright 2017 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.

from yams import ValidationError
from cubicweb.devtools.testlib import CubicWebTC


class SyncSourcesTC(CubicWebTC):
    def test_source_type_unknown(self):
        with self.admin_access.cnx() as cnx:
            with self.assertRaises(ValidationError) as cm:
                cnx.create_entity(
                    "CWSource",
                    name="source",
                    type="doesnotexit",
                    parser="doestnotmatter",
                )
        self.assertIn("Unknown source type", str(cm.exception))

    def test_cant_delete_system_source(self):
        with self.admin_access.cnx() as cnx:
            with self.assertRaises(ValidationError) as cm:
                cnx.execute("DELETE CWSource X")
        self.assertIn("You cannot remove the system source", str(cm.exception))

    def test_cant_rename_system_source(self):
        with self.admin_access.cnx() as cnx:
            with self.assertRaises(ValidationError) as cm:
                cnx.find("CWSource").one().cw_set(name="sexy name")
        self.assertIn("You cannot rename the system source", str(cm.exception))

    def test_cant_add_config_system_source(self):
        with self.admin_access.cnx() as cnx:
            source = cnx.find("CWSource").one()

            with self.assertRaises(ValidationError) as cm:
                source.cw_set(url="whatever")
            self.assertIn(
                "Configuration of the system source goes to the 'sources' file",
                str(cm.exception),
            )

            with self.assertRaises(ValidationError) as cm:
                source.cw_set(config="whatever")
            self.assertIn(
                "Configuration of the system source goes to the 'sources' file",
                str(cm.exception),
            )


if __name__ == "__main__":
    import unittest

    unittest.main()



# copyright 2003-2014 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.

from yams import ValidationError
from cubicweb.devtools.testlib import CubicWebTC
from cubicweb.server.session import Connection

from cubicweb.server.sources.native import UndoTransactionException, _UndoException

from cubicweb.transaction import NoSuchTransaction


class UndoableTransactionTC(CubicWebTC):
    def setup_database(self):
        with self.admin_access.repo_cnx() as cnx:
            self.totoeid = self.create_user(
                cnx, "toto", password="toto", groups=("users",), commit=False
            ).eid
            self.txuuid = cnx.commit()

    def toto(self, cnx):
        return cnx.entity_from_eid(self.totoeid)

    def setUp(self):
        # Force undo feature to be turned on
        Connection.undo_actions = property(lambda self: True, lambda self, v: None)
        super().setUp()

    def tearDown(self):
        super().tearDown()
        del Connection.undo_actions

    def check_transaction_deleted(self, cnx, txuuid):
        # also check transaction actions have been properly deleted
        cu = cnx.system_sql(f"SELECT * from tx_entity_actions WHERE tx_uuid='{txuuid}'")
        self.assertFalse(cu.fetchall())
        cu = cnx.system_sql(
            f"SELECT * from tx_relation_actions WHERE tx_uuid='{txuuid}'"
        )
        self.assertFalse(cu.fetchall())

    def assertUndoTransaction(self, cnx, txuuid, expected_errors=None):
        if expected_errors is None:
            expected_errors = []
        try:
            cnx.undo_transaction(txuuid)
        except UndoTransactionException as exn:
            errors = exn.errors
        else:
            errors = []
        self.assertEqual(errors, expected_errors)

    def test_undo_api(self):
        self.assertTrue(self.txuuid)
        # test transaction api
        with self.admin_access.client_cnx() as cnx:
            tx_actions = cnx.transaction_actions(self.txuuid)
            self.assertEqual(len(tx_actions), 2, tx_actions)
            self.assertRaises(NoSuchTransaction, cnx.transaction_info, "hop")
            self.assertRaises(NoSuchTransaction, cnx.transaction_actions, "hop")
            self.assertRaises(NoSuchTransaction, cnx.undo_transaction, "hop")
            txinfo = cnx.transaction_info(self.txuuid)
            self.assertTrue(txinfo.datetime)
            self.assertEqual(txinfo.user_eid, cnx.user.eid)
            self.assertEqual(txinfo.user().login, "admin")
            actions = txinfo.actions_list()
            self.assertEqual(len(actions), 2)
            actions = txinfo.actions_list(public=False)
            self.assertEqual(len(actions), 6)
            a1 = actions[0]
            self.assertEqual(a1.action, "C")
            self.assertEqual(a1.eid, self.totoeid)
            self.assertEqual(a1.etype, "CWUser")
            self.assertEqual(a1.ertype, "CWUser")
            self.assertEqual(a1.changes, None)
            self.assertEqual(a1.public, True)
            self.assertEqual(a1.order, 1)
            a4 = actions[3]
            self.assertEqual(a4.action, "A")
            self.assertEqual(a4.rtype, "in_group")
            self.assertEqual(a4.ertype, "in_group")
            self.assertEqual(a4.eid_from, self.totoeid)
            self.assertEqual(a4.eid_to, self.toto(cnx).in_group[0].eid)
            self.assertEqual(a4.order, 4)
            for i, rtype in ((1, "owned_by"), (2, "owned_by")):
                a = actions[i]
                self.assertEqual(a.action, "A")
                self.assertEqual(a.eid_from, self.totoeid)
                self.assertEqual(a.rtype, rtype)
                self.assertEqual(a.order, i + 1)
            self.assertEqual(
                {actions[4].rtype, actions[5].rtype},
                {"in_state", "created_by"},
            )
            for i in (4, 5):
                a = actions[i]
                self.assertEqual(a.action, "A")
                self.assertEqual(a.eid_from, self.totoeid)
                self.assertEqual(a.order, i + 1)

            # test undoable_transactions
            txs = cnx.undoable_transactions()
            self.assertEqual(len(txs), 1)
            self.assertEqual(txs[0].uuid, self.txuuid)
            # test transaction_info / undoable_transactions security
        with self.new_access("anon").client_cnx() as cnx:
            self.assertRaises(NoSuchTransaction, cnx.transaction_info, self.txuuid)
            self.assertRaises(NoSuchTransaction, cnx.transaction_actions, self.txuuid)
            self.assertRaises(NoSuchTransaction, cnx.undo_transaction, self.txuuid)
            txs = cnx.undoable_transactions()
            self.assertEqual(len(txs), 0)

    def test_undoable_transactions(self):
        with self.admin_access.client_cnx() as cnx:
            toto = self.toto(cnx)
            cnx.create_entity(
                "EmailAddress", address="toto@logilab.org", reverse_use_email=toto
            )
            txuuid1 = cnx.commit()
            toto.cw_delete()
            txuuid2 = cnx.commit()
            undoable_transactions = cnx.undoable_transactions
            txs = undoable_transactions(action="D")
            self.assertEqual(len(txs), 1, txs)
            self.assertEqual(txs[0].uuid, txuuid2)
            txs = undoable_transactions(action="C")
            self.assertEqual(len(txs), 2, txs)
            self.assertEqual(txs[0].uuid, txuuid1)
            self.assertEqual(txs[1].uuid, self.txuuid)
            txs = undoable_transactions(eid=toto.eid)
            self.assertEqual(len(txs), 3)
            self.assertEqual(txs[0].uuid, txuuid2)
            self.assertEqual(txs[1].uuid, txuuid1)
            self.assertEqual(txs[2].uuid, self.txuuid)
            txs = undoable_transactions(etype="CWUser")
            self.assertEqual(len(txs), 2)
            txs = undoable_transactions(etype="CWUser", action="C")
            self.assertEqual(len(txs), 1)
            self.assertEqual(txs[0].uuid, self.txuuid)
            txs = undoable_transactions(etype="EmailAddress", action="D")
            self.assertEqual(len(txs), 0)
            txs = undoable_transactions(etype="EmailAddress", action="D", public=False)
            self.assertEqual(len(txs), 1)
            self.assertEqual(txs[0].uuid, txuuid2)
            txs = undoable_transactions(eid=toto.eid, action="R", public=False)
            self.assertEqual(len(txs), 1)
            self.assertEqual(txs[0].uuid, txuuid2)

    def test_undo_deletion_base(self):
        with self.admin_access.client_cnx() as cnx:
            toto = self.toto(cnx)
            e = cnx.create_entity(
                "EmailAddress", address="toto@logilab.org", reverse_use_email=toto
            )
            # entity with inlined relation
            cnx.create_entity(
                "CWProperty",
                pkey="ui.default-text-format",
                value="text/rest",
                for_user=toto,
            )
            cnx.commit()
            txs = cnx.undoable_transactions()
            self.assertEqual(len(txs), 2)
            toto.cw_delete()
            txuuid = cnx.commit()
            actions = cnx.transaction_info(txuuid).actions_list()
            self.assertEqual(len(actions), 1)
            toto.cw_clear_all_caches()
            e.cw_clear_all_caches()
            self.assertUndoTransaction(cnx, txuuid)
            undotxuuid = cnx.commit()
            self.assertEqual(undotxuuid, None)  # undo not undoable
            self.assertTrue(cnx.execute("Any X WHERE X eid %(x)s", {"x": toto.eid}))
            self.assertTrue(cnx.execute("Any X WHERE X eid %(x)s", {"x": e.eid}))
            self.assertTrue(cnx.execute('Any X WHERE X has_text "toto@logilab"'))
            self.assertEqual(toto.cw_adapt_to("IWorkflowable").state, "activated")
            self.assertEqual(
                toto.cw_adapt_to("IEmailable").get_email(), "toto@logilab.org"
            )
            self.assertEqual(
                [(p.pkey, p.value) for p in toto.reverse_for_user],
                [("ui.default-text-format", "text/rest")],
            )
            self.assertEqual([g.name for g in toto.in_group], ["users"])
            self.assertEqual(
                [et.name for et in toto.related("is", entities=True)], ["CWUser"]
            )
            self.assertEqual([et.name for et in toto.is_instance_of], ["CWUser"])
            # undoing shouldn't be visble in undoable transaction, and the undone
            # transaction should be removed
            txs = cnx.undoable_transactions()
            self.assertEqual(len(txs), 2)
            self.assertRaises(NoSuchTransaction, cnx.transaction_info, txuuid)
        with self.admin_access.repo_cnx() as cnx:
            self.check_transaction_deleted(cnx, txuuid)
            # the final test: check we can login with the previously deleted user
        with self.new_access("toto").client_cnx():
            pass

    def test_undo_deletion_integrity_1(self):
        with self.admin_access.client_cnx() as cnx:
            # 'Personne fiche Card with' '??' cardinality
            c = cnx.create_entity("Card", title="hop", content="hop")
            p = cnx.create_entity("Personne", nom="louis", fiche=c)
            cnx.commit()
            c.cw_delete()
            txuuid = cnx.commit()
            c2 = cnx.create_entity("Card", title="hip", content="hip")
            p.cw_set(fiche=c2)
            cnx.commit()
            self.assertUndoTransaction(
                cnx,
                txuuid,
                [
                    "Can't restore object relation fiche to entity "
                    "%s which is already linked using this relation." % p.eid
                ],
            )
            cnx.commit()
            p.cw_clear_all_caches()
            self.assertEqual(p.fiche[0].eid, c2.eid)
            # we restored the card
            self.assertTrue(cnx.entity_from_eid(c.eid))

    def test_undo_deletion_integrity_2(self):
        with self.admin_access.client_cnx() as cnx:
            # test validation error raised if we can't restore a required relation
            g = cnx.create_entity("CWGroup", name="staff")
            cnx.execute("DELETE U in_group G WHERE U eid %(x)s", {"x": self.totoeid})
            self.toto(cnx).cw_set(in_group=g)
            cnx.commit()
            self.toto(cnx).cw_delete()
            txuuid = cnx.commit()
            g.cw_delete()
            cnx.commit()
            self.assertUndoTransaction(
                cnx,
                txuuid,
                [
                    "Can't restore relation in_group, object entity "
                    "%s doesn't exist anymore." % g.eid
                ],
            )
            with self.assertRaises(ValidationError) as cm:
                cnx.commit()
            cm.exception.translate(str)
            self.assertEqual(cm.exception.entity, self.totoeid)
            self.assertEqual(
                cm.exception.errors,
                {
                    "in_group-subject": (
                        "at least one relation in_group is "
                        "required on CWUser (%s)" % self.totoeid
                    )
                },
            )

    def test_undo_creation_1(self):
        with self.admin_access.client_cnx() as cnx:
            c = cnx.create_entity("Card", title="hop", content="hop")
            p = cnx.create_entity("Personne", nom="louis", fiche=c)
            txuuid = cnx.commit()
            self.assertUndoTransaction(cnx, txuuid)
            cnx.commit()
            self.assertFalse(cnx.execute("Any X WHERE X eid %(x)s", {"x": c.eid}))
            self.assertFalse(cnx.execute("Any X WHERE X eid %(x)s", {"x": p.eid}))
            self.assertFalse(cnx.execute("Any X,Y WHERE X fiche Y"))
        with self.admin_access.repo_cnx() as cnx:
            for eid in (p.eid, c.eid):
                self.assertFalse(
                    cnx.system_sql(f"SELECT * FROM entities WHERE eid={eid}").fetchall()
                )
                self.assertFalse(
                    cnx.system_sql(
                        f"SELECT 1 FROM owned_by_relation WHERE eid_from={eid}"
                    ).fetchall()
                )
                # added by sql in hooks (except when using dataimport)
                self.assertFalse(
                    cnx.system_sql(
                        f"SELECT 1 FROM is_relation WHERE eid_from={eid}"
                    ).fetchall()
                )
                self.assertFalse(
                    cnx.system_sql(
                        f"SELECT 1 FROM is_instance_of_relation WHERE eid_from={eid}"
                    ).fetchall()
                )
            self.check_transaction_deleted(cnx, txuuid)

    def test_undo_creation_integrity_1(self):
        with self.admin_access.client_cnx() as cnx:
            tutu = self.create_user(cnx, "tutu", commit=False)
            txuuid = cnx.commit()
            email = cnx.create_entity("EmailAddress", address="tutu@cubicweb.org")
            prop = cnx.create_entity(
                "CWProperty", pkey="ui.default-text-format", value="text/html"
            )
            tutu.cw_set(use_email=email, reverse_for_user=prop)
            cnx.commit()
            with self.assertRaises(ValidationError) as cm:
                cnx.undo_transaction(txuuid)
            self.assertEqual(cm.exception.entity, tutu.eid)
            self.assertEqual(
                cm.exception.errors,
                {None: "some later transaction(s) touch entity, undo them first"},
            )

    def test_undo_creation_integrity_2(self):
        with self.admin_access.client_cnx() as cnx:
            g = cnx.create_entity("CWGroup", name="staff")
            txuuid = cnx.commit()
            cnx.execute("DELETE U in_group G WHERE U eid %(x)s", {"x": self.totoeid})
            self.toto(cnx).cw_set(in_group=g)
            cnx.commit()
            with self.assertRaises(ValidationError) as cm:
                cnx.undo_transaction(txuuid)
            self.assertEqual(cm.exception.entity, g.eid)
            self.assertEqual(
                cm.exception.errors,
                {None: "some later transaction(s) touch entity, undo them first"},
            )
        # self.assertEqual(errors,
        #                   [u"Can't restore relation in_group, object entity "
        #                   "%s doesn't exist anymore." % g.eid])
        # with self.assertRaises(ValidationError) as cm: cnx.commit()
        # self.assertEqual(cm.exception.entity, self.totoeid)
        # self.assertEqual(cm.exception.errors,
        #                   {'in_group-subject': u'at least one relation in_group is '
        #                    'required on CWUser (%s)' % self.totoeid})

    # test implicit 'replacement' of an inlined relation

    def test_undo_inline_rel_remove_ok(self):
        """Undo remove relation  Personne (?) fiche (?) Card

        NB: processed by `_undo_r` as expected"""
        with self.admin_access.client_cnx() as cnx:
            c = cnx.create_entity("Card", title="hop", content="hop")
            p = cnx.create_entity("Personne", nom="louis", fiche=c)
            cnx.commit()
            p.cw_set(fiche=None)
            txuuid = cnx.commit()
            self.assertUndoTransaction(cnx, txuuid)
            cnx.commit()
            p.cw_clear_all_caches()
            self.assertEqual(p.fiche[0].eid, c.eid)

    def test_undo_inline_rel_remove_ko(self):
        """Restore an inlined relation to a deleted entity, with an error.

        NB: processed by `_undo_r` as expected"""
        with self.admin_access.client_cnx() as cnx:
            c = cnx.create_entity("Card", title="hop", content="hop")
            p = cnx.create_entity("Personne", nom="louis", fiche=c)
            cnx.commit()
            p.cw_set(fiche=None)
            txuuid = cnx.commit()
            c.cw_delete()
            cnx.commit()
            self.assertUndoTransaction(
                cnx,
                txuuid,
                [
                    "Can't restore relation fiche, object entity %d doesn't exist anymore."
                    % c.eid
                ],
            )
            cnx.commit()
            p.cw_clear_all_caches()
            self.assertFalse(p.fiche)
        with self.admin_access.repo_cnx() as cnx:
            self.assertIsNone(
                cnx.system_sql(
                    f"SELECT cw_fiche FROM cw_Personne WHERE cw_eid={p.eid}"
                ).fetchall()[0][0]
            )

    def test_undo_inline_rel_add_ok(self):
        """Undo add relation  Personne (?) fiche (?) Card

        Caution processed by `_undo_u`, not `_undo_a` !"""
        with self.admin_access.client_cnx() as cnx:
            c = cnx.create_entity("Card", title="hop", content="hop")
            p = cnx.create_entity("Personne", nom="louis")
            cnx.commit()
            p.cw_set(fiche=c)
            txuuid = cnx.commit()
            self.assertUndoTransaction(cnx, txuuid)
            cnx.commit()
            p.cw_clear_all_caches()
            self.assertFalse(p.fiche)

    def test_undo_inline_rel_delete_ko(self):
        with self.admin_access.client_cnx() as cnx:
            c = cnx.create_entity("Card", title="hop", content="hop")
            txuuid = cnx.commit()
            cnx.create_entity("Personne", nom="louis", fiche=c)
            cnx.commit()
            integrityerror = (
                self.repo.system_source.dbhelper.dbapi_module.IntegrityError
            )
            with self.assertRaises(integrityerror):
                cnx.undo_transaction(txuuid)

    def test_undo_inline_rel_add_ko(self):
        """Undo add relation  Personne (?) fiche (?) Card

        Caution processed by `_undo_u`, not `_undo_a` !"""
        with self.admin_access.client_cnx() as cnx:
            c = cnx.create_entity("Card", title="hop", content="hop")
            p = cnx.create_entity("Personne", nom="louis")
            cnx.commit()
            p.cw_set(fiche=c)
            txuuid = cnx.commit()
            c.cw_delete()
            cnx.commit()
            self.assertUndoTransaction(cnx, txuuid)

    def test_undo_inline_rel_replace_ok(self):
        """Undo changing relation  Personne (?) fiche (?) Card

        Caution processed by `_undo_u`"""
        with self.admin_access.client_cnx() as cnx:
            c1 = cnx.create_entity("Card", title="hop", content="hop")
            c2 = cnx.create_entity("Card", title="hip", content="hip")
            p = cnx.create_entity("Personne", nom="louis", fiche=c1)
            cnx.commit()
            p.cw_set(fiche=c2)
            txuuid = cnx.commit()
            self.assertUndoTransaction(cnx, txuuid)
            cnx.commit()
            p.cw_clear_all_caches()
            self.assertEqual(p.fiche[0].eid, c1.eid)

    def test_undo_inline_rel_replace_ko(self):
        """Undo changing relation  Personne (?) fiche (?) Card, with an error

        Caution processed by `_undo_u`"""
        with self.admin_access.client_cnx() as cnx:
            c1 = cnx.create_entity("Card", title="hop", content="hop")
            c2 = cnx.create_entity("Card", title="hip", content="hip")
            p = cnx.create_entity("Personne", nom="louis", fiche=c1)
            cnx.commit()
            p.cw_set(fiche=c2)
            txuuid = cnx.commit()
            c1.cw_delete()
            cnx.commit()
            self.assertUndoTransaction(
                cnx,
                txuuid,
                [
                    "can't restore entity %s of type Personne, target of fiche (eid %s)"
                    " does not exist any longer" % (p.eid, c1.eid)
                ],
            )
            cnx.commit()
            p.cw_clear_all_caches()
            self.assertFalse(p.fiche)

    def test_undo_attr_update_ok(self):
        with self.admin_access.client_cnx() as cnx:
            p = cnx.create_entity("Personne", nom="toto")
            cnx.commit()
            p.cw_set(nom="titi")
            txuuid = cnx.commit()
            self.assertUndoTransaction(cnx, txuuid)
            p.cw_clear_all_caches()
            self.assertEqual(p.nom, "toto")

    def test_undo_attr_update_ko(self):
        with self.admin_access.client_cnx() as cnx:
            p = cnx.create_entity("Personne", nom="toto")
            cnx.commit()
            p.cw_set(nom="titi")
            txuuid = cnx.commit()
            p.cw_delete()
            cnx.commit()
            self.assertUndoTransaction(
                cnx,
                txuuid,
                [
                    "can't restore state of entity %s, it has been deleted inbetween"
                    % p.eid
                ],
            )


class UndoExceptionInUnicode(CubicWebTC):
    # problem occurs in string manipulation for python < 2.6
    def test___str__method(self):
        u = _UndoException("voilà")
        self.assertIsInstance(str(u), str)


if __name__ == "__main__":
    from logilab.common.testlib import unittest_main

    unittest_main()



# copyright 2003-2010 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""mime type transformation engine for cubicweb, based on mtconverter"""


from logilab import mtconverter
from logilab.mtconverter import (
    register_base_transforms,
    register_pil_transforms,
    register_pygments_transforms,
)
from logilab.mtconverter.engine import TransformEngine
from logilab.mtconverter.transform import Transform

from cubicweb.ext.markdown import markdown_publish
from cubicweb.uilib import rest_publish, html_publish
from cubicweb import schema


HTML_MIMETYPES = ("text/html", "text/xhtml", "application/xhtml+xml")

# CubicWeb specific transformations


class rest_to_html(Transform):
    inputs = ("text/rest", "text/x-rst")
    output = "text/html"

    def _convert(self, trdata):
        return rest_publish(trdata.appobject, trdata.decode())


class markdown_to_html(Transform):
    inputs = ("text/markdown", "text/x-markdown")
    output = "text/html"

    def _convert(self, trdata):
        return markdown_publish(trdata.appobject, trdata.decode())


class html_to_html(Transform):
    inputs = HTML_MIMETYPES
    output = "text/html"

    def _convert(self, trdata):
        return html_publish(trdata.appobject, trdata.data)


# Instantiate and configure the transformation engine

mtconverter.UNICODE_POLICY = "replace"

ENGINE = TransformEngine()
ENGINE.add_transform(rest_to_html())
ENGINE.add_transform(markdown_to_html())
ENGINE.add_transform(html_to_html())

schema.NEED_PERM_FORMATS.remove("text/cubicweb-page-template")


if register_pil_transforms(ENGINE, verb=False):
    HAS_PIL_TRANSFORMS = True
else:
    HAS_PIL_TRANSFORMS = False

try:
    from logilab.mtconverter.transforms import pygmentstransforms

    for mt in ("text/plain",) + HTML_MIMETYPES:
        try:
            pygmentstransforms.mimetypes.remove(mt)
        except ValueError:
            continue
    register_pygments_transforms(ENGINE, verb=False)

    def patch_convert(cls):
        def _convert(self, trdata, origconvert=cls._convert):
            add_css = getattr(trdata.appobject._cw, "add_css", None)
            if add_css is not None:
                # session has no add_css, only http request
                add_css("pygments.css")
            return origconvert(self, trdata)

        cls._convert = _convert

    patch_convert(pygmentstransforms.PygmentsHTMLTransform)

    HAS_PYGMENTS_TRANSFORMS = True
except ImportError:
    HAS_PYGMENTS_TRANSFORMS = False

register_base_transforms(ENGINE, verb=False)
